import os, sys

import h5py
import numpy as np

sys.path.append("\\".join(os.getcwd().split("\\")[:-1]))
import logging
import select

import socket
import time

from abc import ABC

import pythoncom

from epics import caget, caget_many, caput
from win32com import client


class AcceleratorInterface(ABC):
    """
    low level controller_interface class used to communicate with the accelerator
    - should impliment the following via overwriting
        - __init__() method to establish connections to control computer
        - set_beamline() method to send PV's to pyEPICS et. al.
    - due to the complex nature of custom observations_list please define a specific observation class to
      get measurements, feel free to define methods here to do so
    """

    def __init__(self):
        """establish connections here"""
        self.logger = logging.getLogger(__name__)

    def set_parameters(self, params, pvals):
        """
        set beamline PV's here
        Arguments
        ---------
        params : list
            List of parameter objects (see parameter.py)
        pvals : ndarray
            Array of parameter set points (unnormalized)
        """
        raise NotImplementedError

    def get_parameters(self, params):
        raise NotImplementedError


class AWAInterface(AcceleratorInterface):
    """
    controller_interface class to connect to AWA control system

    """

    UseNIFG = False
    Initialized = False
    m_AWACameraHost = "127.0.0.1"
    m_AWAPGCamPort = 2019
    m_AWANIFGPort = 2029
    NewMeasurement = False
    FWHMX = 10000
    FMHMY = 10000
    FMHML = 10000
    TempVal = 0.0
    Testing = False

    #    USBDIO = client.Dispatch('USBDIOCtrl.Application')

    def __init__(self, use_frame_grabber=False, testing=False):

        super().__init__()
        self.ni_frame_grabber = client.Dispatch("NIFGCtrl")
        self.AWAPGCamera = client.Dispatch("AWAPGCamera.application")
        self.m_CameraClient = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        self.logger = logging.getLogger(__name__)
        self.logger.info("Starting controller_interface")

        self.Testing = testing

        self.initialize_connections(use_frame_grabber)
        self.logger.info("Done")

    def initialize_connections(self, use_frame_grabber):
        # image client

        if self.Testing:
            self.Initialized = False
        else:
            if self.Initialized:
                self.logger.debug("Deleteing old objects")
                del self.ni_frame_grabber
                del self.AWAPGCamera
                self.m_CameraClient.close()
                self.m_CameraClient = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            else:
                pass

            pythoncom.CoInitialize()
            self.UseNIFG = use_frame_grabber
            if not self.UseNIFG:
                self.logger.debug("Connecting to AWAPGCamera application")
                self.m_CameraPort = self.m_AWAPGCamPort
            else:
                self.logger.debug("Connecting to NI Frame Grabber application")
                self.m_CameraPort = self.m_AWANIFGPort
            self.m_CameraClient.connect(("127.0.0.1", self.m_CameraPort))
            self.m_CameraClient.setblocking(0)

            self.Initialized = True
        # setpoint client
        # self.AWABeamlineClient = socket.socket(socket.AF_INET, socket. SOCK_STREAM)

    def close(self):
        if not self.Initialized:
            return

        if self.UseNIFG:
            del self.ni_frame_grabber

        else:
            del self.AWAPGCamera

    def get_raw_image(self):
        if self.Initialized:
            if self.UseNIFG:
                return self.ni_frame_grabber.GetImage()
            else:
                return self.AWAPGCamera.GetImage
        else:
            raise RuntimeError(
                "Trying to retrieve an image before controller_interface is initialized!"
            )

    def get_roi(self):
        if self.Initialized:
            if self.UseNIFG:
                module = self.ni_frame_grabber
            else:
                module = self.AWAPGCamera

            x1 = module.ROIX1
            x2 = module.ROIX2
            y1 = module.ROIY1
            y2 = module.ROIY2

            return np.array(((x1, y1), (x2, y2)))
        else:
            raise RuntimeError(
                "Trying to retrieve an image "
                "before controller_interface is initialized!"
            )

    def get_data(
        self, target_charge=-1, charge_deviation=0.1, n_samples=1, stat=False, bpm=True
    ):
        """
        get new image and charge data

        Arguments
        ---------
        target_charge : float, optional
            Target charge for valid observation in nC, if negative ignore.
            Default: -1 (ignore)

        charge_deviation : float, optional
            Fractional deviation from target charge on ICT1 allowed for valid
            observation. Default: 0.1

        n_samples : int
            Number of samples to take

        note - calculations of centroids and FWHM etc. are based on a region of
        interest, which might be changed by the user!

        Connect to camera broadcasting TCP port for notifications
        If it is timed out, then just download and return whatever
        image available
        In order to avoid the complication of TCP buffer cleanup
        we will simply close the connection and reopen it
        elf.
        """

        self.logger.debug(f"taking n samples {n_samples}")

        n_shots = 0
        roi = self.get_roi()
        self.img = []
        self.charge = np.empty(n_samples)
        self.phase = np.empty(n_samples)
        self.bpmx = np.empty(n_samples)
        self.bpmy = np.empty(n_samples)
        self.CentroidY = np.empty(n_samples)
        self.CentroidX = np.empty(n_samples)
        self.FWHML = np.empty(n_samples)
        self.FWHMY = np.empty(n_samples)
        self.FWHMX = np.empty(n_samples)

        # reset trigger to start acuisition
        caput("AWATrigCtrl:Reset", 1)

        if not self.Testing:
            self.logger.debug("restarting camera client")
            self.m_CameraClient.close()
            self.m_CameraClient = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.m_CameraClient.connect(("127.0.0.1", self.m_CameraPort))
            self.m_CameraClient.setblocking(0)

        print("make sure Master input is set to 0 (frame grabber) or (1) PGCam")
        while n_shots < n_samples:
            print("shot", n_shots)
            caput("AWATrigCtrl:GetCh8.PROC", 1)
            while caget("AWATrigCtrl:GetCh8") > 0:
                caput("AWATrigCtrl:GetCh8.PROC", 1)

            if not self.Testing:
                ready = select.select([self.m_CameraClient], [], [], 2)

                if ready[0]:
                    # gate measurement
                    # self.USBDIO.SetReadyState(8, 0)

                    # check charge on ICT1 is within bounds or charge bounds is not
                    # specified (target charge < 0)
                    ict1_charge = np.abs(caget("AWAVXI11ICT:Ch1"))
                    a = self.m_CameraClient.recv(1024)
                    if (
                        np.abs(ict1_charge - target_charge)
                        < np.abs(charge_deviation * target_charge)
                    ) or (target_charge < 0):

                        # a = self.m_CameraClient.recv(1024)
                        # print(a)
                        b = "".join(chr(x) for x in a)
                        if stat:
                            try:
                                c = eval(b)
                                #
                                self.FWHMX[n_shots] = c["FWHMX"]
                                self.FWHMY[n_shots] = c["FWHMY"]
                                self.FWHML[n_shots] = c["FWHML"]
                                self.CentroidX[n_shots] = c["CX"]
                                self.CentroidY[n_shots] = c["CY"]
                                self.NewMeasurement = True
                                rawdata = self.get_raw_image()
                            except SyntaxError:
                                self.logger.warning("sleeping!{}".format(b))
                                time.sleep(0.1)
                        if not stat:
                            self.NewMeasurement = True
                            rawdata = self.get_raw_image()

                            # get charge
                        self.charge[n_shots] = ict1_charge
                        # get phase
                        if bpm:  # This is commented: 2022. 07. 06
                            results = caget("AWABPPM:Results")
                            self.phase[n_shots] = results[0]
                            self.bpmx[n_shots] = results[2]
                            self.bpmy[n_shots] = results[3]
                        #                            self.phase[n_shots] = caget('AWABPPM:Ch1')
                        # get bpm
                        #                            self.bpmx[n_shots] = caget('AWABPPM:Ch3')
                        #                            self.bpmy[n_shots] = caget('AWABPPM:Ch4')

                        # get ROI
                        # roi = self.get_roi()
                        # self.logger.debug(roi)

                        n_shots += 1
                        caput("AWATrigCtrl:SetCh8", 1)
                        self.img += [np.array(rawdata)]

                    else:
                        # if we are considering charge limits then print a warning
                        if target_charge > 0:
                            self.logger.warning(
                                f"ICT1 charge:{ict1_charge:.3e} nC"
                                f" is outside target range"
                            )
                            time.sleep(0.1)

                        #                    self.USBDIO.SetReadyState(8, 1)
                        caput("AWATrigCtrl:SetCh8", 1)

                else:
                    self.logger.warning("camera client not ready for data")

            else:
                # generate testing data
                img, data = self.get_test_data()
                self.img += [img]
                self.FWHMX[n_shots] = data
                self.FWHMY[n_shots] = data
                self.FWHML[n_shots] = data
                self.CentroidX[n_shots] = data
                self.CentroidY[n_shots] = data

                roi = np.array(((0, 0), (700, 700)))
                self.charge[n_shots] = np.ones(4)
                n_shots += 1

        self.img = np.asarray(self.img)

        # collect scalar data
        sdata = np.c_[
            self.FWHMX,
            self.FWHMY,
            self.FWHML,
            self.CentroidX,
            self.CentroidY,
            self.bpmx,
            self.bpmy,
            self.phase,
            self.charge,
        ]

        return self.img, sdata, roi

    def save_data(
        self,
        fname,
        target_charge=-1,
        charge_deviation=0.1,
        n_samples=1,
        slit_scan=False,
        bpm_window=None,
        stat=False,
    ):
        assert not os.path.exists(f"{fname}.hdf5"), "File already existed"
        self.img = []
        self.charge = np.empty(n_samples)
        self.phase = np.empty(n_samples)
        self.bpmx = np.empty(n_samples)
        self.bpmy = np.empty(n_samples)

        self.img, sdata, roi = self.get_data(
            target_charge, charge_deviation, n_samples, stat
        )

        self.charge = sdata[:, -1]
        self.phase = sdata[:, -2]
        self.bpmy = sdata[:, -3]
        self.bpmx = sdata[:, -4]

        # bpm and phase window
        window_bpmx = (self.bpmx > self.bpmx.mean() - 0.8 * self.bpmx.std()) & (
            self.bpmx < self.bpmx.mean() + 0.8 * self.bpmx.std()
        )
        window_bpmy = (self.bpmy > self.bpmy.mean() - 0.8 * self.bpmy.std()) & (
            self.bpmy < self.bpmy.mean() + 0.8 * self.bpmy.std()
        )
        window_phase = (self.phase > self.phase.mean() - 5) & (
            self.phase < self.phase.mean() + 5
        )

        if not slit_scan:
            window = np.ones(n_samples, dtype=bool)
            print("no bpm or phase window")
        elif slit_scan == "x":
            window = window_bpmx
        elif slit_scan == "y":
            window = window_bpmy
        elif slit_scan == "xy":
            window = window_bpmx & window_bpmy
        elif slit_scan == "phase":
            window = window_phase
        else:
            raise ValueError

        #        if bpm_window:
        #           bpmx_mean, bpmx_std = bpm_window[0][0], bpm_window[0][1]
        #           bpmy_mean, bpmy_std = bpm_window[1][0], bpm_window[1][1]
        #           window_bpmx = (self.bpmx > bpmx_mean - bpmx_std) & (self.bpmx < bpmx_mean + bpmx_std)
        #           window_bpmy = (self.bpmy > bpmy_mean - bpmy_std) & (self.bpmy < bpmy_mean + bpmy_std)
        #           window = window_bpmx & window_bpmy
        #       else:
        #           window = np.ones(n_samples, dtype=bool)

        # remove window on 08-26-2022 PP due to boolean mismatch error.
        self.charge = self.charge[window]
        self.phase = self.phase[window]
        self.bpmy = self.bpmy[window]
        self.bpmx = self.bpmx[window]
        self.img = self.img[window]

        print("- " * 20)
        if len(self.charge) == 0:
            raise ValueError("no shot in window")
        else:
            if slit_scan:
                print(f"{len(self.charge)}/{n_samples} shots in window")

        with h5py.File(f"{fname}.hdf5", "w") as f:
            # dataset
            d_charge = f.create_dataset("charge", data=self.charge, compression="lzf")
            d_image = f.create_dataset("image", data=self.img, compression="lzf")
            d_phase = f.create_dataset("phase", data=self.phase, compression="lzf")
            d_bpmx = f.create_dataset("bpmx", data=self.bpmx, compression="lzf")
            d_bpmy = f.create_dataset("bpmy", data=self.bpmy, compression="lzf")
            # attrs
            d_charge.attrs["target_charge"] = target_charge
            d_image.attrs["roi"] = roi
            print(f"saving to {fname}.hdf5")

        #         with h5py.File(f'{fname}.hdf5', 'w') as f:
        #             d_charge = f.create_group("charge")
        #             d_image = f.create_group("image")
        #             d_phase = f.create_group("phase")
        #             d_bpmx = f.create_group("bpmx")
        #             d_bpmy = f.create_group("bpmy")
        #             f['/'].attrs['roi'] = roi
        #             f['/'].attrs['n_samples'] = len(self.charge)
        #             f['/'].attrs['target_charge'] = target_charge
        #             print(f'saving to {fname}.hdf5 ...')
        #             for i in range(len(self.charge)):
        #                 d_charge.create_dataset(f'charge_{i}',data=self.charge[i])
        #                 d_phase.create_dataset(f'phase_{i}',data=self.phase[i])
        #                 d_image.create_dataset(f'image_{i}',data=self.img[i])
        #                 d_bpmx.create_dataset(f'bpmx_{i}',data=self.bpmx[i])
        #                 d_bpmy.create_dataset(f'bpmy_{i}',data=self.bpmy[i])
        #             print('saved!')
        # collect scalar data
        return self.img, sdata, roi

    def set_parameters(self, parameters, pvvals):
        assert len(parameters) == len(pvvals)
        self.TempVal = pvvals

        if self.Testing:
            return

        for i in range(len(parameters)):
            self.logger.debug("sending epics command")
            self.logger.debug(f"caput {parameters[i].channel} {pvvals[i]}")
            status = caput(parameters[i].channel, pvvals[i])
            self.logger.debug(f"caput return status {status}")

    def get_parameters(self, parameters):
        if self.Testing:
            return np.random.rand(len(parameters))

        channels = [ele.channel for ele in parameters]
        vals = caget_many(channels)
        return np.array(vals)

    def get_test_data(self):
        r = np.sin(self.TempVal)
        tmp = sum(r)
        data = np.genfromtxt("test_images/onenc250200.txt", names=True)

        size = 50e-3
        bins = 700
        img, xedges, yedges = np.histogram2d(
            data["x"],
            data["y"],
            range=np.array(((-0.0, 1.0), (-0.0, 1.0))) * size / 2.0,
            bins=(bins, bins),
        )

        img = img / np.max(img)
        img = 0.15 * np.random.rand(*img.shape) + img

        return img, tmp


class LinacPhaseController:
    def __init__(self, linac, phase0=0):
        assert linac in ["L1", "L2", "L3", "L4", "L5", "L6"]
        self.phase0 = phase0
        self.linac = linac
        self.history = []

    def set_abs_phase(self, phase):
        p0 = self.get_abs_phase()
        dp = phase - p0
        dp_temp = self._set_rel_phase(dp)
        print(f"d_phase: {dp}, {dp_temp}")
        self.history.append(dp)

    def get_abs_phase(self):
        if len(self.history):
            return sum(np.asarray(self.history))
        else:
            return 0

    def get_steps(self):
        RBCh = {
            "L1": "SMCtrl:AWA:pos03",
            "L2": "SMCtrl:AWA:pos13",
            "L3": "SMCtrl:AWA:pos14",
            "L4": "SMCtrl:AWA:pos11",
            "L5": "AWALLRF:K3:MeasuredPhase",
            "L6": "AWALLRF:K4:MeasuredPhase",
        }

        return caget(RBCh[self.linac])

    def _set_rel_phase(self, phase):

        controller = {
            "L1": self.L1Delay,
            "L2": self.L2Delay,
            "L3": self.L3Delay,
            "L4": self.L4Delay,
            "L5": self.L5Delay,
            "L6": self.L6Delay,
        }

        return controller[self.linac](phase)

    def K2Delay(self, delta):
        CtrlCh = "SMCtrl:AWA:Rel03"
        RBCh = "SMCtrl:AWA:pos03"
        StepsPerDegree = 64
        StepstoMove = delta * StepsPerDegree
        oldpos = caget(RBCh)
        caput(CtrlCh, StepstoMove)
        time.sleep(15)
        newpos = caget(RBCh)
        print(oldpos)
        print(newpos)
        return delta

    def L2Delay(self, delta):
        CtrlCh = "SMCtrl:AWA:Rel13"
        RBCh = "SMCtrl:AWA:pos13"
        StepsPerDegree = 881
        StepstoMove = delta * StepsPerDegree
        oldpos = caget(RBCh)
        caput(CtrlCh, StepstoMove)
        time.sleep(15)
        newpos = caget(RBCh)
        print(oldpos)
        print(newpos)
        return delta

    def L3Delay(self, delta):
        CtrlCh = "SMCtrl:AWA:Rel14"
        RBCh = "SMCtrl:AWA:pos14"
        StepsPerDegree = -881
        StepstoMove = delta * StepsPerDegree
        oldpos = caget(RBCh)
        caput(CtrlCh, StepstoMove)
        time.sleep(15)
        newpos = caget(RBCh)
        print(oldpos)
        print(newpos)
        return delta

    def L4Delay(self, delta):
        CtrlCh = "SMCtrl:AWA:Rel11"
        RBCh = "SMCtrl:AWA:pos11"
        StepsPerDegree = -881
        StepstoMove = delta * StepsPerDegree
        oldpos = caget(RBCh)
        caput(CtrlCh, StepstoMove)
        time.sleep(15)
        newpos = caget(RBCh)
        print(oldpos)
        print(newpos)
        return delta

    def K3Delay(self, delta):
        CtrlCh = "AWALLRF:K3:PLLTarget"
        RBCh = "AWALLRF:K3:MeasuredPhase"
        oldpos = caget(RBCh)
        caput(CtrlCh, delta + oldpos)
        time.sleep(15)
        newpos = caget(RBCh)
        print(oldpos)
        print(newpos)
        return delta

    def K4Delay(self, delta):
        CtrlCh = "AWALLRF:K4:PLLTarget"
        RBCh = "AWALLRF:K4:MeasuredPhase"
        oldpos = caget(RBCh)
        caput(CtrlCh, delta + oldpos)
        time.sleep(15)
        newpos = caget(RBCh)
        print(oldpos)
        print(newpos)
        return delta

    def L1Delay(self, delta):
        self.K2Delay(delta)
        self.L2Delay(-delta)
        return delta

    def L5Delay(self, delta):
        self.K3Delay(delta)
        self.L3Delay(-delta)
        return delta

    def L6Delay(self, delta):
        self.K4Delay(delta)
        self.L4Delay(-delta)
        return delta
